@@ -1,37 +1,56 @@
 from bloodyAD import utils, asciitree
 from bloodyAD.exceptions import LOG
-from bloodyAD.network.ldap import Scope
+from bloodyAD.network.ldap import Scope, showRecoverable
 from bloodyAD.exceptions import NoResultError
-from bloodyAD.formatters import common
-from msldap.commons.exceptions import LDAPSearchException
+from bloodyAD.formatters import bloodhound as bhformatter
+from badldap.commons.exceptions import LDAPSearchException
 from typing import Literal
-import re, asyncio
+import re
 
 
-def children(conn, target: str = "DOMAIN", otype: str = "*", direct: bool = False):
+async def bloodhound(conn, transitive: bool = False, path: str = "CurrentPath"):
+    """
+    BloodHound CE collector (WARNING: This script is still in development. It only provides the basics - ADCS ESC and other complex nodes aren't supported yet)
+
+    :param transitive: if set, will try to reach trusts to have more complete results (you should start from a dc of your user domain to have more complete results)
+    :param path: filepath for the generated zip file
+    """
+    from badldap import bloodhound
+
+    output_path = None
+    if path != "CurrentPath":
+        output_path = path
+    ldap = await conn.getLdap()
+    bh = bloodhound.MSLDAPDump2Bloodhound(ldap.co_url, follow_trusts=transitive, output_path=output_path)
+    await bh.run()
+
+async def children(conn, target: str = "DOMAIN", otype: str = "*", direct: bool = False):
     """
     List children for a given target object
 
-    :param target: sAMAccountName, DN, GUID or SID of the target
-    :param otype: special keyword "useronly" or objectClass of objects to fetch e.g. user, computer, group, trustedDomain, organizationalUnit, container, groupPolicyContainer, msDS-GroupManagedServiceAccount, etc
+    :param target: sAMAccountName, DN or SID of the target
+    :param otype: special keyword "useronly" or objectClass of objects to fetch (e.g. computer, group, trustedDomain, organizationalUnit, container, groupPolicyContainer, msDS-GroupManagedServiceAccount, etc)
     :param direct: Fetch only direct children of target
     """
+    ldap = await conn.getLdap()
     if target == "DOMAIN":
-        target = conn.ldap.domainNC
+        target = ldap.domainNC
     scope = Scope.LEVEL if direct else Scope.SUBTREE
     if otype == "useronly":
         otype_filter = "sAMAccountType=805306368"
     else:
         otype_filter = f"objectClass={otype}"
-    return conn.ldap.bloodysearch(
+    async for entry in ldap.bloodysearch(
         target,
         f"(&({otype_filter})(!(distinguishedName={target})))",
         search_scope=scope,
         attr=["distinguishedName"],
-    )
+        controls=showRecoverable()
+    ):
+        yield entry
 
 
-def dnsDump(conn, zone: str = None, no_detail: bool = False, transitive: bool = False):
+async def dnsDump(conn, zone: str = None, no_detail: bool = False, transitive: bool = False):
     """
     Retrieve DNS records of the Active Directory readable/listable by the user
 
@@ -40,7 +59,7 @@ def dnsDump(conn, zone: str = None, no_detail: bool = False, transitive: bool =
     :param transitive: if set, try to fetch dns records in AD trusts (you should start from a DC of your user domain to have exhaustive results)
     """
 
-    def domainDnsDump(conn, zone=None, no_detail=False):
+    async def domainDnsDump(conn, zone=None, no_detail=False):
         entries = None
         filter = "(|(objectClass=dnsNode)(objectClass=dnsZone))"
         prefix_blacklist = [
@@ -62,16 +81,17 @@ def dnsDump(conn, zone: str = None, no_detail: bool = False, transitive: bool =
                 prefix_filter += f"(!(name={prefix}))"
             filter = f"(&{filter}{prefix_filter})"
 
+        ldap = await conn.getLdap()
         dnsZones = []
-        for nc in conn.ldap.appNCs + [conn.ldap.domainNC]:
+        for nc in ldap.appNCs + [ldap.domainNC]:
             try:
-                entries = conn.ldap.bloodysearch(
+                entries = ldap.bloodysearch(
                     nc,
                     filter,
                     search_scope=Scope.SUBTREE,
                     attr=["dnsRecord", "name", "objectClass"],
                 )
-                for entry in entries:
+                async for entry in entries:
                     domain_suffix = entry["distinguishedName"].split(",")[1]
                     domain_suffix = domain_suffix.split("=")[1]
 
@@ -143,29 +163,29 @@ def dnsDump(conn, zone: str = None, no_detail: bool = False, transitive: bool =
                                     }
                                 )
                         except KeyError:
-                            LOG.error("[-] KeyError for record: " + record)
+                            LOG.error("KeyError for record: " + record)
                             continue
 
                     yield yield_entry
             except (NoResultError, LDAPSearchException) as e:
                 if type(e) is NoResultError:
-                    LOG.warning(f"[!] No readable record found in {nc}")
+                    LOG.warning(f"No readable record found in {nc}")
                 else:
-                    LOG.warning(f"[!] {nc} couldn't be read on {conn.conf.host}")
+                    LOG.warning(f"{nc} couldn't be read on {conn.conf.host}")
                 continue
         # List record names if we have list child right on dnsZone or MicrosoftDNS container but no READ_PROP on record object
-        for nc in conn.ldap.appNCs:
+        for nc in ldap.appNCs:
             dnsZones.append(f"CN=MicrosoftDNS,{nc}")
-        dnsZones.append(f"CN=MicrosoftDNS,CN=System,{conn.ldap.domainNC}")
+        dnsZones.append(f"CN=MicrosoftDNS,CN=System,{ldap.domainNC}")
         for searchbase in dnsZones:
             try:
-                entries = conn.ldap.bloodysearch(
+                entries = ldap.bloodysearch(
                     searchbase,
                     "(objectClass=*)",
                     search_scope=Scope.SUBTREE,
                     attr=["objectClass"],
                 )
-                for entry in entries:
+                async for entry in entries:
                     # If we can get objectClass it means we have a READ_PROP on the record object so we already found it before
                     if (
                         entry.get("objectClass")
@@ -195,16 +215,17 @@ def dnsDump(conn, zone: str = None, no_detail: bool = False, transitive: bool =
                     yield {"recordName": domain_name}
             except (NoResultError, LDAPSearchException) as e:
                 if type(e) is NoResultError:
-                    LOG.warning(f"[!] No listable record found in {nc}")
+                    LOG.warning(f"No listable record found in {nc}")
                 else:
-                    LOG.warning(f"[!] {nc} couldn't be read on {conn.conf.host}")
+                    LOG.warning(f"{nc} couldn't be read on {conn.conf.host}")
                 continue
 
     # Used to avoid duplicate entries if there is the same record in multiple partitions
     record_dict = {}
     record_entries = []
+    ldap = await conn.getLdap()
     if transitive:
-        trustmap = conn.ldap.getTrustMap()
+        trustmap = await ldap.getTrustMap()
         for trust in trustmap.values():
             if "conn" in trust:
                 record_entries.append(domainDnsDump(trust["conn"], zone, no_detail))
@@ -213,7 +234,7 @@ def dnsDump(conn, zone: str = None, no_detail: bool = False, transitive: bool =
 
     basic_records = []
     for records in record_entries:
-        for r in records:
+        async for r in records:
             keyName = "recordName" if "recordName" in r else "zoneName"
             # If it's a record with only the record name returns it later and only if we didn't find another record with more info
             if len(r) == 1 and keyName == "recordName":
@@ -251,63 +272,63 @@ def dnsDump(conn, zone: str = None, no_detail: bool = False, transitive: bool =
         yield basic_r
 
 
-def membership(conn, target: str, no_recurse: bool = False):
+async def membership(conn, target: str, no_recurse: bool = False):
     """
     Retrieve SID and SAM Account Names of all groups a target belongs to
 
-    :param target: sAMAccountName, DN, GUID or SID of the target
+    :param target: sAMAccountName, DN or SID of the target
     :param no_recurse: if set, doesn't retrieve groups where target isn't a direct member
     """
+    ldap = await conn.getLdap()
     ldap_filter = ""
     if no_recurse:
-        entries = conn.ldap.bloodysearch(target, attr=["objectSid", "memberOf"])
-        for entry in entries:
+        entries = ldap.bloodysearch(target, attr=["objectSid", "memberOf"])
+        async for entry in entries:
             for group in entry.get("memberOf", []):
                 ldap_filter += f"(distinguishedName={group})"
         if not ldap_filter:
-            LOG.warning("[!] No direct group membership found")
-            return []
+            LOG.warning("No direct group membership found")
+            return
     else:
         # [MS-ADTS] 3.1.1.4.5.19 tokenGroups, tokenGroupsNoGCAcceptable
         attr = "tokenGroups"
-        entries = conn.ldap.bloodysearch(target, attr=[attr])
-        for entry in entries:
+        entries = ldap.bloodysearch(target, attr=[attr])
+        async for entry in entries:
             try:
                 for groupSID in entry[attr]:
                     ldap_filter += f"(objectSID={groupSID})"
             except KeyError:
-                LOG.warning("[!] No membership found")
-                return []
+                LOG.warning("No membership found")
+                return
         if not ldap_filter:
             LOG.warning("no GC Server available, the set of groups might be incomplete")
             attr = "tokenGroupsNoGCAcceptable"
-            entries = conn.ldap.bloodysearch(target, attr=[attr])
-            for entry in entries:
+            entries = ldap.bloodysearch(target, attr=[attr])
+            async for entry in entries:
                 for groupSID in entry[attr]:
                     ldap_filter += f"(objectSID={groupSID})"
 
-    entries = conn.ldap.bloodysearch(
-        conn.ldap.domainNC,
+    entries = ldap.bloodysearch(
+        ldap.domainNC,
         f"(|{ldap_filter})",
         search_scope=Scope.SUBTREE,
         attr=["objectSID", "sAMAccountName"],
     )
-    return entries
+    async for entry in entries:
+        yield entry
 
 
-def trusts(conn, transitive: bool = False):
+async def trusts(conn, transitive: bool = False):
     """
     Display trusts in an ascii tree starting from the DC domain as tree root. A->B means A can auth on B and A-<B means B can auth on A, A-<>B means bidirectional
 
     :param transitive: Try to fetch transitive trusts (you should start from a dc of your user domain to have more complete results)
     """
-
-    trust_dict = asyncio.get_event_loop().run_until_complete(
-        conn.ldap.getTrusts(transitive, conn.conf.dns)
-    )
+    ldap = await conn.getLdap()
+    trust_dict = await ldap.getTrusts(transitive, conn.conf.dns)
 
     # Get the host domain as root for the trust tree
-    trust_root_domain = (".".join(conn.ldap.domainNC.split(",DC="))).split("DC=")[1]
+    trust_root_domain = ldap.domainname
     if trust_dict:
         tree = {}
         asciitree.branchFactory({":" + trust_root_domain: tree}, [], trust_dict)
@@ -315,45 +336,56 @@ def trusts(conn, transitive: bool = False):
         print(tree_printer({trust_root_domain: tree}))
 
 
-def object(
-    conn, target: str, attr: str = "*", resolve_sd: bool = False, raw: bool = False
+async def object(
+    conn,
+    target: str,
+    attr: str = "*",
+    resolve_sd: bool = False,
+    raw: bool = False,
+    transitive: bool = False,
 ):
     """
     Retrieve LDAP attributes for the target object provided, binary data will be outputted in base64
 
-    :param target: sAMAccountName, DN, GUID or SID of the target (if you give an empty string "" prints rootDSE)
+    :param target: sAMAccountName, DN or SID of the target (if you give an empty string "" prints rootDSE)
     :param attr: attributes to retrieve separated by a comma, retrieves all the attributes by default
     :param resolve_sd: if set, permissions linked to a security descriptor will be resolved (see bloodyAD github wiki/Access-Control for more information)
     :param raw: if set, will return attributes as sent by the server without any formatting, binary data will be outputted in base64
+    :param transitive: if set with "--resolve-sd", will try to resolve foreign SID by reaching trusts
     """
     attributesSD = [
         "nTSecurityDescriptor",
         "msDS-GroupMSAMembership",
         "msDS-AllowedToActOnBehalfOfOtherIdentity",
     ]
-    entries = conn.ldap.bloodysearch(target, attr=attr.split(","), raw=raw)
+    conn.conf.transitive = transitive
+    ldap = await conn.getLdap()
+    entries = ldap.bloodysearch(target, attr=attr.split(","), raw=raw)
     rendered_entries = utils.renderSearchResult(entries)
     if resolve_sd and not raw:
-        for entry in rendered_entries:
+        async for entry in rendered_entries:
             for attrSD in attributesSD:
                 if attrSD in entry:
                     e = entry[attrSD]
                     if not isinstance(e, list):
-                        entry[attrSD] = utils.renderSD(e, conn)
+                        entry[attrSD] = await utils.renderSD(e, conn)
                     else:
-                        entry[attrSD] = [utils.renderSD(sd, conn) for sd in e]
+                        entry[attrSD] = [await utils.renderSD(sd, conn) for sd in e]
             yield entry
     else:
-        yield from rendered_entries
+        async for entry in rendered_entries:
+            yield entry
 
 
-def search(
+async def search(
     conn,
     base: str = "DOMAIN",
     filter: str = "(objectClass=*)",
     attr: str = "*",
     resolve_sd: bool = False,
     raw: bool = False,
+    transitive: bool = False,
+    c: list = [],
 ):
     """
     Search in LDAP database, binary data will be outputted in base64
@@ -363,58 +395,69 @@ def search(
     :param attr: attributes to retrieve separated by a comma
     :param resolve_sd: if set, permissions linked to a security descriptor will be resolved (see bloodyAD github wiki/Access-Control for more information)
     :param raw: if set, will return attributes as sent by the server without any formatting, binary data will be outputed in base64
+    :param transitive: if set with "--resolve-sd", will try to resolve foreign SID by reaching trusts
+    :param c: if set, will use the controls for extended search operations, e.g. "-c 1.2.840.113556.1.4.2064 -c 1.2.840.113556.1.4.2065" to display tombstoned, deleted and recycled objects and their linked attributes
     """
     attributesSD = [
         "nTSecurityDescriptor",
         "msDS-GroupMSAMembership",
         "msDS-AllowedToActOnBehalfOfOtherIdentity",
     ]
+    conn.conf.transitive = transitive
+    ldap = await conn.getLdap()
     if base == "DOMAIN":
-        base = conn.ldap.domainNC
-    entries = conn.ldap.bloodysearch(
-        base, filter, search_scope=Scope.SUBTREE, attr=attr.split(","), raw=raw
+        base = ldap.domainNC
+    # RFC2251 4.1.12 Controls
+    # control ::= (controlType, criticality, controlValue)
+    controls = [(oid,True,None) for oid in c]
+    entries = ldap.bloodysearch(
+        base, filter, search_scope=Scope.SUBTREE, attr=attr.split(","), raw=raw, controls=controls
     )
     rendered_entries = utils.renderSearchResult(entries)
     if resolve_sd and not raw:
-        for entry in rendered_entries:
+        async for entry in rendered_entries:
             for attrSD in attributesSD:
                 if attrSD in entry:
                     e = entry[attrSD]
                     if not isinstance(e, list):
-                        entry[attrSD] = utils.renderSD(e, conn)
+                        entry[attrSD] = await utils.renderSD(e, conn)
                     else:
-                        entry[attrSD] = [utils.renderSD(sd, conn) for sd in e]
+                        entry[attrSD] = [await utils.renderSD(sd, conn) for sd in e]
             yield entry
     else:
-        yield from rendered_entries
+        async for entry in rendered_entries:
+            yield entry
 
 
-# TODO: Search writable for application partitions too?
-def writable(
+async def writable(
     conn,
-    otype: Literal["ALL", "OU", "USER", "COMPUTER", "GROUP", "DOMAIN", "GPO"] = "ALL",
+    otype: str = "*",
     right: Literal["ALL", "WRITE", "CHILD"] = "ALL",
     detail: bool = False,
-    # partition: Literal["DOMAIN", "DNS", "ALL"] = "DOMAIN"
+    partition: Literal["DOMAIN", "CONFIGURATION", "SCHEMA", "DNS", "ALL"] = "ALL",
+    transitive: bool = False,
+    exclude_del: bool = False,
+    bh: bool = False,
 ):
     """
     Retrieve objects writable by client
 
-    :param otype: type of writable object to retrieve
+    :param otype: special keywords "useronly, ou, gpo" or objectClass of objects to fetch (e.g. computer, group, server...)
     :param right: type of right to search
     :param detail: if set, displays attributes/object types you can write/create for the object
+    :param partition: directory partition a.k.a naming context to explore
+    :param transitive: if set, will try to reach trusts to have more complete results (you should start from a dc of your user's domain to have more complete results)
+    :param exclude_del: if set, exclude deleted objects
+    :param bh: if set, creates a BloodHound-compatible Zip file with the writable objects found
     """
-    # :param partition: directory partition a.k.a naming context to explore
 
     ldap_filter = ""
-    if otype == "USER":
+    if otype == "useronly":
         ldap_filter = "(sAMAccountType=805306368)"
+    elif otype == "ou":
+        ldap_filter = "(|(objectClass=container)(objectClass=organizationalUnit))"
     else:
-        if otype == "ALL":
-            objectClass = "*"
-        elif otype == "OU":
-            objectClass = "container"
-        elif otype == "GPO":
+        if otype == "gpo":
             objectClass = "groupPolicyContainer"
         else:
             objectClass = otype
@@ -451,32 +494,64 @@ def writable(
             "right": "CREATE_CHILD",
         }
 
-    searchbases = []
-    # if partition == "DOMAIN":
-    searchbases.append(conn.ldap.domainNC)
-    # elif partition == "DNS":
-    #     searchbases.append(conn.ldap.applicationNCs) # A definir https://learn.microsoft.com/en-us/windows/win32/ad/enumerating-application-directory-partitions-in-a-forest
-    # else:
-    #     searchbases.append(conn.ldap.NCs) # A definir
-    right_entry = {}
-    for searchbase in searchbases:
-        for entry in conn.ldap.bloodysearch(
-            searchbase, ldap_filter, search_scope=Scope.SUBTREE, attr=attr_params.keys()
-        ):
-            for attr_name in entry:
-                if attr_name not in attr_params:
-                    continue
-                key_names = attr_params[attr_name]["lambda"](entry[attr_name])
-                for name in key_names:
-                    if name == "distinguishedName":
-                        name = "dn"
-                    if name not in right_entry:
-                        right_entry[name] = []
-                    right_entry[name].append(attr_params[attr_name]["right"])
+    controls = None
+    if not exclude_del:
+        controls = showRecoverable()
+    
+    async def search_writable(conn):
+        ldap = await conn.getLdap()
+        searchbases = []
+        if partition == "DOMAIN":
+            searchbases.append(ldap.domainNC)
+        elif partition == "CONFIGURATION":
+            searchbases.append(ldap.configNC)
+        elif partition == "SCHEMA":
+            searchbases.append(ldap.schemaNC)
+        elif partition == "DNS":
+            searchbases = ldap.applicationNCs
+        else: # For ALL option
+            searchbases = ldap._serverinfo["namingContexts"]
+        
+        right_entry = {}
+        for searchbase in searchbases:
+            writable_entries = []
+            async for entry in ldap.bloodysearch(
+                searchbase, ldap_filter, search_scope=Scope.SUBTREE, attr=attr_params, controls=controls
+            ):
+                for attr_name in entry:
+                    if attr_name not in attr_params:
+                        continue
+                    key_names = attr_params[attr_name]["lambda"](entry[attr_name])
+                    for name in key_names:
+                        if name == "distinguishedName":
+                            name = "dn"
+                        if name not in right_entry:
+                            right_entry[name] = []
+                        right_entry[name].append(attr_params[attr_name]["right"])
 
-            if right_entry:
-                yield {
-                    **{"distinguishedName": entry["distinguishedName"]},
-                    **right_entry,
-                }
-                right_entry = {}
+                if right_entry:
+                    # Build base result with distinguishedName
+                    result = {"distinguishedName": entry["distinguishedName"]}        
+                    if bh:
+                        writable_entries.append(entry["distinguishedName"])        
+                    # Merge right_entry into result
+                    result.update(right_entry)         
+                    yield result
+                    right_entry = {}
+            if bh:
+                current_user_dn = await ldap.dnResolver(ldap.creds.username)
+                current_user_groups = [g['distinguishedName'] async for g in membership(conn, current_user_dn)]
+                writable_entries += [current_user_dn] + current_user_groups
+                bh_filter = "(|" + "".join(f"(distinguishedName={utils.double_encode_controls(dn)})" for dn in writable_entries) + ")"
+                await bhformatter.granular_bh(conn, searchbase, bh_filter)
+
+    root_ldap = await conn.getLdap()
+    if transitive:
+        trustmap = await root_ldap.getTrustMap()
+        for trust in trustmap.values():
+            if "conn" in trust:
+                async for result in search_writable(trust["conn"]):
+                    yield result
+    else:
+        async for result in search_writable(conn):
+            yield result
\ No newline at end of file
